@@ -66,7 +66,7 @@ BEGIN {
 	}) {
 		@ISA = qw(IO::Socket::INET);
 	}
-	$VERSION = '1.16_3';
+	$VERSION = '1.30_2';
 	$GLOBAL_CONTEXT_ARGS = {};
 
 	#Make $DEBUG another name for $Net::SSLeay::trace
@@ -209,8 +209,6 @@ sub configure_SSL {
 	my %default_args = (
 		Proto => 'tcp',
 		SSL_server => $is_server,
-		SSL_ca_file => 'certs/my-ca.pem',
-		SSL_ca_path => 'ca/',
 		SSL_use_cert => $is_server,
 		SSL_check_crl => 0,
 		SSL_version	=> 'sslv23',
@@ -231,45 +229,48 @@ sub configure_SSL {
 			:  "certs/client-${k}.pem";
 	}	
 
+	# add only SSL_ca_* if not in args
+	if ( ! exists $arg_hash->{SSL_ca_file} && ! exists $arg_hash->{SSL_ca_path} ) {
+		if ( -f 'certs/my-ca.pem' ) {
+			$default_args{SSL_ca_file} = 'certs/my-ca.pem'
+		} elsif ( -d 'ca/' ) {
+			$default_args{SSL_ca_path} = 'ca/'
+		}
+	}
+		
 	#Replace nonexistent entries with defaults
 	%$arg_hash = ( %default_args, %$GLOBAL_CONTEXT_ARGS, %$arg_hash );
 
 	#Avoid passing undef arguments to Net::SSLeay
 	defined($arg_hash->{$_}) or delete($arg_hash->{$_}) foreach (keys %$arg_hash);
 
-	#Handle CA paths properly if no CA file is specified
-	if ($arg_hash->{'SSL_ca_path'} ne '' and !(-f $arg_hash->{'SSL_ca_file'})) {
-		DEBUG(1, "CA file $arg_hash->{SSL_ca_file} not found, using CA path instead.\n" )
-			if $arg_hash->{SSL_ca_file};
-		$arg_hash->{'SSL_ca_file'} = '';
-	}
-
 	my $vcn_scheme = delete $arg_hash->{SSL_verifycn_scheme};
 	if ( $vcn_scheme && $vcn_scheme ne 'none' ) {
+		# don't access ${*self} inside callback - this seems to create
+		# circular references from the ssl object to the context and back
+
+		# use SSL_verifycn_name or determine from PeerAddr
+		my $host = $arg_hash->{SSL_verifycn_name};
+		if (not defined($host)) {
+			if ( $host = $arg_hash->{PeerAddr} || $arg_hash->{PeerHost} ) {
+				$host =~s{:[a-zA-Z0-9_\-]+$}{};
+			}
+		}
+		$host ||= ref($vcn_scheme) && $vcn_scheme->{callback} && 'unknown';
+		$host or return $self->error( "Cannot determine peer hostname for verification" );
+
 		my $vcb = $arg_hash->{SSL_verify_callback};
 		$arg_hash->{SSL_verify_callback} = sub {
-			my ($ok,$ctx_store,$cert,$error) = @_;
-			$ok = $vcb->($ok,$ctx_store,$cert,$error) if $vcb;
+			my ($ok,$ctx_store,$certname,$error,$cert) = @_;
+			$ok = $vcb->($ok,$ctx_store,$certname,$error,$cert) if $vcb;
 			$ok or return;
 			my $depth = Net::SSLeay::X509_STORE_CTX_get_error_depth($ctx_store);
 			return $ok if $depth != 0;
 
-			# use SSL_peer_hostname or determine from PeerAddr
-			my $arg_hash = ${*$self}{_SSL_arguments};
-			my $host = $arg_hash->{SSL_verifycn_name};
-			if (not defined($host)) {
-				$host = ( $arg_hash->{PeerAddr} || $arg_hash->{PeerHost} );
-				$host =~s{:\w+$}{} if ! $host;
-			}
-			$host ||= ref($vcn_scheme) && $vcn_scheme->{callback} && 'unknown';
-			$host or return $self->error( "Cannot determine peer hostname for verification" );
-
 			# verify name
-			my $x509 = Net::SSLeay::X509_STORE_CTX_get_current_cert($ctx_store);
-			my $rv = verify_hostname_of_cert( $host,$x509,$vcn_scheme );
+			my $rv = verify_hostname_of_cert( $host,$cert,$vcn_scheme );
 			# just do some code here against optimization because x509 has no
 			# increased reference and CRYPTO_add is not available from Net::SSLeay
-			DEBUG(99999,"don't to anything with $x509" );
 			return $rv;
 		};
 	}
@@ -644,11 +645,14 @@ sub generic_write {
 	my $written;
 	if ( $write_all ) {
 		my $data = $length < $buf_len-$offset ? substr($$buffer, $offset, $length) : $$buffer;
-		$written = Net::SSLeay::ssl_write_all($ssl, $data);
+		($written, my $errs) = Net::SSLeay::ssl_write_all($ssl, $data);
+		# ssl_write_all returns number of bytes written
+		$written = undef if ! $written && $errs;
 	} else {
 		$written = Net::SSLeay::write_partial( $ssl,$offset,$length,$$buffer );
+		# write_partial does SSL_write which returns -1 on error
+		$written = undef if $written < 0;
 	}
-	$written = undef if $written < 0; # Net::SSLeay::write returns -1 not undef on error
 	if ( !defined($written) ) {
 		$self->_set_rw_error( $ssl,-1 )
 			|| $self->error("SSL write error");
@@ -783,8 +787,8 @@ sub close {
 sub stop_SSL {
 	my $self = shift || return _invalid_object();
 	my $stop_args = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
-	return $self->error("SSL object already closed") 
-		unless (${*$self}{'_SSL_opened'} == 1);
+	return $self->error("SSL object not open") 
+		if ! ${*$self}{'_SSL_opened'};
 
 	if (my $ssl = ${*$self}{'_SSL_object'}) {
 		my $shutdown_done;
@@ -858,7 +862,7 @@ sub stop_SSL {
 sub kill_socket {
 	my $self = shift;
 	shutdown($self, 2);
-	$self->close(SSL_no_shutdown => 1) if (${*$self}{'_SSL_opened'} == 1);
+	$self->close(SSL_no_shutdown => 1) if ${*$self}{'_SSL_opened'};
 	delete(${*$self}{'_SSL_ctx'});
 	return;
 }
@@ -947,8 +951,8 @@ sub dump_peer_certificate {
 
 {
 	my %dispatcher = (
-		issuer =>	  sub { Net::SSLeay::X509_NAME_oneline( Net::SSLeay::X509_get_issuer_name( shift )) },
-		subject =>	  sub { Net::SSLeay::X509_NAME_oneline( Net::SSLeay::X509_get_subject_name( shift )) },
+		issuer =>  sub { Net::SSLeay::X509_NAME_oneline( Net::SSLeay::X509_get_issuer_name( shift )) },
+		subject => sub { Net::SSLeay::X509_NAME_oneline( Net::SSLeay::X509_get_subject_name( shift )) },
 	);
 	if ( $Net::SSLeay::VERSION >= 1.30 ) {
 		# I think X509_NAME_get_text_by_NID got added in 1.30
@@ -1036,6 +1040,7 @@ sub dump_peer_certificate {
 	$scheme{imap} = $scheme{ldap}; # rfc 2595
 	$scheme{acap} = $scheme{ldap}; # rfc 2595
 	$scheme{nntp} = $scheme{ldap}; # rfc 4642
+	$scheme{ftp}  = $scheme{http}; # rfc 4217
 
 	# function to verify the hostname
 	#
@@ -1075,8 +1080,9 @@ sub dump_peer_certificate {
 			 # definitly no hostname, try IPv4
 			$ip4 = inet_aton( $identity ) or croak "'$identity' is not IPv4, but neither IPv6 nor hostname";
 		} else {
-			# assume hostname
-			if ( $identity !~m{^[a-zA-Z0-9-_\.]+$} ) {
+			# assume hostname, check for umlauts etc
+			if ( $identity =~m{[^a-zA-Z0-9_.\-]} ) {
+				$identity =~m{\0} and croak("name '$identity' has \\0 byte");
 				$identity = idn_to_ascii($identity) or
 					croak "Warning: Given name '$identity' could not be converted to IDNA!";
 			}
@@ -1093,12 +1099,12 @@ sub dump_peer_certificate {
 			# The RFCs are in this regard unspecific but we don't want to have to
 			# deal with certificates like *.com, *.co.uk or even *
 			# see also http://nils.toedtmann.net/pub/subjectAltName.txt
-			if ( $wtyp eq 'anywhere' and $name =~m{^([\w\-]*)\*(.+)} ) {
-				$pattern = qr{^\Q$1\E[\w\-]*\Q$2\E$}i;
+			if ( $wtyp eq 'anywhere' and $name =~m{^([a-zA-Z0-9_\-]*)\*(.+)} ) {
+				$pattern = qr{^\Q$1\E[a-zA-Z0-9_\-]*\Q$2\E$}i;
 			} elsif ( $wtyp eq 'leftmost' and $name =~m{^\*(\..+)$} ) {
-				$pattern = qr{^[\w\-]*\Q$1\E$}i;
+				$pattern = qr{^[a-zA-Z0-9_\-]*\Q$1\E$}i;
 			} else {
-				$pattern = qr{^\Q$name}i;
+				$pattern = qr{^\Q$name\E$}i;
 			}
 			return $identity =~ $pattern;
 		};
@@ -1124,7 +1130,7 @@ sub dump_peer_certificate {
 
 		if ( $scheme->{check_cn} eq 'always' or 
 			$scheme->{check_cn} eq 'when_only' and !$alt_dnsNames) {
-			$check_name->($commonName,$identity,$scheme->{wildcars_in_cn})
+			$check_name->($commonName,$identity,$scheme->{wildcards_in_cn})
 				and return 1;
 		}
 
@@ -1182,7 +1188,7 @@ sub error {
 sub DESTROY {
 	my $self = shift || return;
 	$self->close(_SSL_in_DESTROY => 1, SSL_no_shutdown => 1) 
-		if (${*$self}{'_SSL_opened'} == 1);
+		if ${*$self}{'_SSL_opened'};
 	delete(${*$self}{'_SSL_ctx'});
 }
 
@@ -1291,6 +1297,7 @@ sub CLOSE {							 #<---- Do not change this function!
 
 
 package IO::Socket::SSL::SSL_Context;
+use Carp;
 use strict;
 
 my %CTX_CREATED_IN_THIS_THREAD;
@@ -1340,10 +1347,15 @@ sub new {
 
 
 	my $verify_mode = $arg_hash->{SSL_verify_mode};
-	unless ($verify_mode == Net::SSLeay::VERIFY_NONE()) {
-		Net::SSLeay::CTX_load_verify_locations(
-			$ctx, $arg_hash->{SSL_ca_file},$arg_hash->{SSL_ca_path}
-		) || return IO::Socket::SSL->error("Invalid certificate authority locations");
+	if ( $verify_mode != Net::SSLeay::VERIFY_NONE() and 
+		! Net::SSLeay::CTX_load_verify_locations( 
+			$ctx, $arg_hash->{SSL_ca_file} || '',$arg_hash->{SSL_ca_path} || '') ) {
+		if ( ! $arg_hash->{SSL_ca_file} && ! $arg_hash->{SSL_ca_path} ) {
+			carp("No certificate verification because neither SSL_ca_file nor SSL_ca_path known");
+			$verify_mode = Net::SSLeay::VERIFY_NONE();
+		} else {
+			return IO::Socket::SSL->error("Invalid certificate authority locations");
+		}
 	}
 
 	if ($arg_hash->{'SSL_check_crl'}) {
@@ -1409,21 +1421,22 @@ sub new {
 	my $verify_cb = $arg_hash->{SSL_verify_callback};
 	my $verify_callback = $verify_cb && sub {
 		my ($ok, $ctx_store) = @_;
-		my ($cert, $error);
+		my ($certname,$cert,$error);
 		if ($ctx_store) {
 			$cert = Net::SSLeay::X509_STORE_CTX_get_current_cert($ctx_store);
 			$error = Net::SSLeay::X509_STORE_CTX_get_error($ctx_store);
-			$cert &&= Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_issuer_name($cert)).
+			$certname = Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_issuer_name($cert)).
 				Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_subject_name($cert));
 			$error &&= Net::SSLeay::ERR_error_string($error);
 		}
 		DEBUG(3, "ok=$ok cert=$cert" );
-		return $verify_cb->($ok, $ctx_store, $cert, $error);
+		return $verify_cb->($ok,$ctx_store,$certname,$error,$cert);
 	};
 
 	Net::SSLeay::CTX_set_verify($ctx, $verify_mode, $verify_callback);
 
 	$ctx_object = { context => $ctx };
+	$ctx_object->{has_verifycb} = 1 if $verify_callback;
 	DEBUG(3, "new ctx $ctx" );
 	$CTX_CREATED_IN_THIS_THREAD{$ctx} = 1;
 
@@ -1462,6 +1475,11 @@ sub DESTROY {
 		DEBUG( 3,"free ctx $ctx open=".join( " ",keys %CTX_CREATED_IN_THIS_THREAD ));
 		if ( %CTX_CREATED_IN_THIS_THREAD and 
 			delete $CTX_CREATED_IN_THIS_THREAD{$ctx} ) {
+			# remove any verify callback for this context
+			if ( $self->{has_verifycb}) {
+				DEBUG( 3,"free ctx $ctx callback" );
+				Net::SSLeay::CTX_set_verify($ctx, 0,undef);
+			}
 			DEBUG( 3,"OK free ctx $ctx" );
 			Net::SSLeay::CTX_free($ctx);
 		}
@@ -1738,7 +1756,7 @@ added.
 
 Specifies session cache object which should be used instead of creating a new.
 Overrules SSL_session_cache_size.
-This option is useful if you wan't to reuse the cache, but not the rest of
+This option is useful if you want to reuse the cache, but not the rest of
 the context.
 
 A session cache object can be created using 
@@ -2100,6 +2118,10 @@ SSL_password_cb.
 IO::Socket::SSL does not work together with Storable::fd_retrieve/fd_store.
 See BUGS file for more information and how to work around the problem.
 
+Non-blocking and timeouts (which are based on non-blocking) are not
+supported on Win32, because the underlying IO::Socket::INET does not support
+non-blocking on this platform.
+
 =head1 LIMITATIONS
 
 IO::Socket::SSL uses Net::SSLeay as the shiny interface to OpenSSL, which is
